import os
import click
import uuid

from laai.hpo.mlflow_utils import terminate_parent_run, get_active_run_feature_set_id, get_active_run_params

VERSION = "0.0.1"
IMAGE_NAME = "osipov/laai"
IMAGE_TAG = "latest"
IMAGE = f"{IMAGE_NAME}:{IMAGE_TAG}"

def assert_aws_credentials():
	assert 'AWS_ACCESS_KEY_ID' in os.environ, "Missing environment variable AWS_ACCESS_KEY_ID"
	assert 'AWS_DEFAULT_REGION' in os.environ, "Missing environment variable AWS_DEFAULT_REGION"
	assert 'AWS_SECRET_ACCESS_KEY' in os.environ, "Missing environment variable AWS_SECRET_ACCESS_KEY"

def prepare_docker():
	import docker
	client = docker.from_env()
	images = client.images.list(IMAGE)
	image = images[0] if len(images) else client.images.pull(IMAGE_NAME, tag = IMAGE_TAG)
	return client


@click.group()
def cli():
	pass

@cli.command("init")
@click.option('--provider', required=True, type=click.Choice(['aws', 'azure', 'gcp'], case_sensitive=False))
@click.option('--backend', default='docker', type=click.Choice(['docker', 'k8s'], case_sensitive=False), show_default=True)
@click.option('--instances', default='1', type=int, show_default=True)
@click.option('--worker-instance-type', default='t3.micro', show_default=True)
@click.option('--manager-instances', default='1', type=int, show_default=True)
@click.option('--manager-instance-type', default='t3.micro', show_default=True)
@click.option('--volume-size', default='8', type=int, show_default=True)
@click.option('--ssh-connection-timeout', default='5', type=int, show_default=True)
@click.option('--ssh-connection-attempts', default='1', type=int, show_default=True)
def init(provider, 
			backend, 
			instances, 
			worker_instance_type, 
			manager_instances, 
			manager_instance_type,
			volume_size,
			ssh_connection_timeout,
			ssh_connection_attempts):			
	"""Initialize a training dojo in a specified infrastructure provider."""
	assert provider == 'aws', f"Provider {provider} is not publicly available in version {VERSION}"
	assert backend == 'docker', f"Backend {backend} is not publicly available in version {VERSION}"

	#check for the aws settings
	if (provider == 'aws'):
		assert_aws_credentials()

		# #drop the provider prefix for the instance type, e.g. aws:t3.micro -> t3.micro
		# assert worker_instance_type[:len(provider) + 1] == provider + ":",\
		# 	f"Worker instance type {worker_instance_type} is missing a prefix {provider}:"
		# worker_instance_type = worker_instance_type[len(provider) + 1:]

		# manager_instance_type = manager_instance_type[len(provider) + 1:]

	print(f"Checking for the latest version...", end='')
	client = prepare_docker()
	print('done')

	dojo = os.environ['DOJO'] if 'DOJO' in os.environ else uuid.uuid4().hex[:8]
	print(f"Using a dojo store at {os.getcwd()}/.dojo/.{dojo}")

	print(f"Preparing a dojo using {backend} on {provider} with {instances} worker(s) and {manager_instances} manager(s) ...")
	container = client.containers.run(	image=IMAGE, 
		command="/bin/bash -c '/opt/laai/dojo_init.sh'",
		volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
											"mode": "rw"}} ,
		environment = { "LAAI_DOJO": dojo, 
						"LAAI_VERSION": VERSION,
						"LAAI_BACKEND": backend,
						"LAAI_PROVIDER": provider,
						"LAAI_DOJO_CONNECT_TIMEOUT": ssh_connection_timeout,
						"LAAI_DOJO_CONNECT_ATTEMPTS": ssh_connection_attempts,
						"AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
						"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
						"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
						"TF_VAR_instances": instances,
						"TF_VAR_worker_instance_type": worker_instance_type,
						"TF_VAR_manager_instances": manager_instances,
						"TF_VAR_manager_instance_type": manager_instance_type,
						"TF_VAR_volume_size": volume_size,
						},
		auto_remove = True,
		detach=True)
		
	logs = container.attach(stream = True, logs = True)
	for s in logs:
		print(s.decode('utf-8').strip())	

	print(dojo)

class DojoCLI(click.MultiCommand):
	@click.command()	
	@click.argument('dojo')
	def activate(dojo):
		assert_aws_credentials()
		print(f"Activating dojo {dojo}")
		
		client = prepare_docker()
		
		container = client.containers.run(	
			image=IMAGE, 
			command="/bin/bash -c '/opt/laai/dojo_activate.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
											"mode": "rw"}} ,
			environment = { "LAAI_DOJO": dojo, 
							# "AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
							# "AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
							# "AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
							
							},
			auto_remove = False,
			detach = True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8'))	

	@click.command()	
	def ls():		
		client = prepare_docker()
		
		container = client.containers.run(	
			image=IMAGE, 
			command="/bin/bash -c '/opt/laai/dojo_ls.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
											"mode": "rw"}} ,
			auto_remove = False,
			detach = True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())	

	@click.command()
	def prune():
		client = prepare_docker()
		container = client.containers.run(	
			image=IMAGE, 
			command="/bin/bash -c '/opt/laai/dojo_prune.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
											"mode": "rw"}} ,
			auto_remove = True,
			detach=True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			s = s.decode('utf-8').strip()
			if len(s):
				print(s)
	
	@click.command()
	@click.argument('dojo')
	def rm(dojo):
		assert_aws_credentials()
		print(f"Removing dojo {dojo} and deactivating store {os.getcwd()}/.dojo/.{dojo}")
		
		client = prepare_docker()
		
		container = client.containers.run(	
			image=IMAGE, 
			command="/bin/bash -c '/opt/laai/dojo_rm.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
											"mode": "rw"}} ,
			environment = { "AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
							"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
							"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
							"LAAI_DOJO": dojo, 
							},
			auto_remove = False,
			detach=True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8'))	
	
	def inspect(self, dojo):
		assert_aws_credentials()
		
		client = prepare_docker()
		
		container = client.containers.run(	
			image=IMAGE, 
			command="/bin/bash -c '/opt/laai/dojo_inspect.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
											"mode": "rw"}} ,
			environment = { "AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
							"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
							"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
							"LAAI_DOJO": dojo, 
							},
			auto_remove = False,
			detach=True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8'))	


	def list_commands(self, ctx):
		return ['ls', 'rm', 'prune', 'activate', 'inspect']

	def get_command(self, ctx, name):
		return {"ls": DojoCLI.ls,
				"rm": DojoCLI.rm,
				"prune": DojoCLI.prune,
				"inspect": DojoCLI.inspect,
				'activate': DojoCLI.activate}[name]

@cli.command(cls=DojoCLI)
def dojo():
	"""Manage a dojo training environment."""
	pass

class JobCLI(click.MultiCommand):
	@click.command()
	@click.option('--image', required = True)
	@click.option('--dojo', required = True)
	@click.option('--subnet', required = False)
	@click.option('--manager-ip', required = False)
	def create(image, dojo, subnet, manager_ip):

		if subnet is not None or manager_ip is not None:
			assert subnet is not None and manager_ip is not None, "Both --subnet and --manager-ip must be specified"
			from ipaddress import ip_network, ip_address
			net = ip_network(subnet)
			assert ip_address(manager_ip) in net, f"Manager IP {manager_ip} is not in the CIDR subnet {subnet} range"

		print(f"Checking for the latest version...", end='')
		client = prepare_docker()
		print('done')

		job = uuid.uuid4().hex[:8]
		print(f"Creating a job {job} using image {image} in dojo {dojo}")

		container = client.containers.run(image = IMAGE, 
			command="/bin/bash -c '/opt/laai/job_create.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
												"mode": "rw"}} ,
			environment = { "LAAI_DOJO": dojo,
							"LAAI_JOB": job,
							"LAAI_JOB_IMAGE": image,
							"LAAI_JOB_SUBNET": subnet if subnet else None,
							"LAAI_JOB_MANAGER_IP": manager_ip if subnet else None,
							},
			auto_remove = True,
			detach=True)
		
		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())	
		
		print(job)			

	@click.command()
	@click.argument('job')
	@click.option('--replicas', default='1', type=int)
	@click.option('--env', '-e', multiple=True, type=click.Tuple([str, str]) )
	def start(job, replicas, env):
		print(f"Checking for the latest version...", end='')
		client = prepare_docker()
		print('done')

		env = dict({ "LAAI_JOB": job,
					 "LAAI_JOB_REPLICAS": replicas},
					 **{"LAAI_JOB_ARG_" + k : v for k, v in env})

		container = client.containers.run(	image = IMAGE, 
			command="/bin/bash -c '/opt/laai/job_start.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
												"mode": "rw"}} ,
			environment = env,
			auto_remove = True,
			detach = True)

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())	

	@click.command()	
	def ls():
		print("run ls")

	@click.command()
	def rm():
		print("run rm")

	def list_commands(self, ctx):
		return ['ls', 'rm', 'create']

	def get_command(self, ctx, name):
		return {"ls": JobCLI.ls,
				"rm": JobCLI.rm,
				'create': JobCLI.create,
				"start": JobCLI.start}[name]

@cli.command("job", cls=JobCLI)
def job():
	pass


class HpoCLI(click.MultiCommand):
	def list_commands(self, ctx):
		return ['start', 'stop', 'set']

	def get_command(self, ctx, name):
		return {"start": HpoCLI.start,
				"stop": HpoCLI.stop,
				"client": HpoCLI.client,
				'set': HpoCLI._set,
				'kv': HpoCLI.kv}[name]

	@click.command()
	@click.argument('name')
	def _set(name):
		print(get_active_run_feature_set_id(feature_set_id = name))

	@click.command()
	@click.option('--output', default='bash', required = False)
	def kv(output):
  		print(get_active_run_params(output = output))

	@click.command()
	@click.option('--num_runs', required = False, default = 1)
	@click.option('--client_prefix', required = False, default = 'laai.hpo.client')
	@click.option('--client_name', required = False, default = 'BaseMLFlowClient')
	def client(num_runs, client_prefix, client_name):
		import importlib.util
		mod = importlib.import_module(client_prefix)
		klass = getattr(mod, client_name)
		client = klass()
		client.run(num_runs = num_runs)

	@click.command()
	@click.option('--name', required = True)
	@click.option('--num_runs', required = False, default = 1)
	@click.option('--hpo_seed', required = False, default = 42)
	@click.option('--service_prefix', required = False, default = 'laai.hpo.services')
	@click.option('--service_name', required = False, default = 'BaseMLFlowService')
	def start(name, num_runs, hpo_seed, service_prefix, service_name):
		import importlib.util
		mod = importlib.import_module(service_prefix)
		klass = getattr(mod, service_name)
		hpo_service = klass()
		hpo_service.run(mlflow_experiment_name = name,
						num_trials = num_runs, 
						hpo_seed = hpo_seed)

	@click.command()
	def stop():
		terminate_parent_run()


@cli.command(cls=HpoCLI)
def hpo():
	"""Manage hyperparameter optimization."""
	pass


# # @click.group(invoke_without_command=False)
# # @click.pass_context
# # def dojo_cmds(ctx):
# # 	pass
# # 	# print("dojo")
# # 	# if ctx.invoked_subcommand is None:
# # 	# 	click.echo('I was invoked without subcommand')
# # 	# else:
# # 	# 	click.echo('I am about to invoke %s' % ctx.invoked_subcommand)

# # @dojo_cmds.command()	
# # @click.pass_context
# # def dojo_ls(ctx):
# # 	print("dojo ls")
# # 	if ctx.invoked_subcommand is None:
# # 		click.echo('I was invoked without subcommand')
# # 	else:
# # 		click.echo('I am about to invoke %s' % ctx.invoked_subcommand)


# # @main.command("dojo")
# # @click.option('ls', required=False)
# # @click.pass_context
# # def dojo_main(ctx):
# # 	print("dojo main")
# # 	# cli = click.CommandCollection(sources=[dojo_cmds])
# # 	# cli(ctx)



# # 	# pass


# # # if __name__ == '__main__':
# # # 	cli = click.CommandCollection(sources=[dojo, run])
# # # 	cli()


# # @click.group(chain=True, invoke_without_command=True)
# # @click.pass_context
# # def dojo_cmds(ctx):
# #     if ctx.invoked_subcommand is None:
# #         click.echo('I was invoked without subcommand')
# #     else:
# #         click.echo('I am about to invoke %s' % ctx.invoked_subcommand)

# # @dojo_cmds.command("dojo")
# # def dojo():
# # 	print("dojo")

# # @dojo_cmds.command("dojo ls")
# # def dojo():
# # 	print("dojo")

# # @dojo_cmds.command()
# # def ls():
# # 	print("ls")

# # @click.group()
# # def run():
# # 	pass

# # @run.command()
# # def run_ls():
# # 	print("run ls")




import os
import argparse
import json 

import time
from datetime import datetime

import numpy as np
import torch as pt
import pandas as pd
import pytorch_lightning as pl

from torch.nn.functional import mse_loss

from mlflow.entities import RunStatus
from laai.hpo.mlflow_utils import get_client, get_parent_run, get_active_child_run, assert_mlflow_tracking_uri, assert_mlflow_experiment_name, get_existing_run
from osds.utils import ObjectStorageDataset
from torch.utils.data import DataLoader

from laai.hpo.client import BaseMLFlowClient

import optuna
from laai.hpo.services import BaseOptunaService
class XorExperimentService(BaseOptunaService):
  def hparams(self):
    trial = self._trial

    #define hyperparameters
    return {
      "seed": trial.suggest_int('seed', 0, np.iinfo(np.int32).max - 1),
        
      "bins": trial.suggest_categorical('bins', [32, 48, 64, 96, 128]),
        
      "optimizer": trial.suggest_categorical('optimizer', ['SGD', 'Adam']),
        
      "lr": trial.suggest_loguniform('lr', 89e-3, 91e-3),
        
      "num_hidden_neurons": [trial.suggest_int(f"num_hidden_layer_{layer}_neurons", 5, 17) \
                                for layer in range(trial.suggest_int('num_layers', 1, 2))],
      
      "batch_size": trial.suggest_categorical('batch_size', [32, 64, 128]),
        
      "max_epochs": trial.suggest_int('max_epochs', 100, 400, log = True)
    }

  def hparams_set_id(self):
    #specify the subsets of hyperparameter name/values to id
    return {
        "hp_bin_id": 'bins'
    }    

  def on_experiment_end(self, experiment, parent_run):
    study = self._study
    try:
      for key, fig in {
        "plot_param_importances": optuna.visualization.plot_param_importances(study),
        "plot_parallel_coordinate": optuna.visualization.plot_parallel_coordinate(study, params=["max_epochs", "num_hidden_layer_0_neurons"]),
        "plot_parallel_coordinate_max_epochs_lr": optuna.visualization.plot_parallel_coordinate(study, params=["max_epochs", "lr"]),
        "plot_parallel_coordinate_lr_layer_0": optuna.visualization.plot_parallel_coordinate(study, params=["lr", "num_hidden_layer_0_neurons"]),
        "plot_contour": optuna.visualization.plot_contour(study, params=["max_epochs", "num_hidden_layer_0_neurons"]),
        "plot_contour_max_epochs_lr": optuna.visualization.plot_contour(study, params=["max_epochs", "lr"]),
      }.items():
        fig.write_image(key + ".png")
        self.mlflow_client.log_artifact(run_id = parent_run.info.run_id, 
                            local_path = key + ".png")
    except:
      print(f"Failed to correctly persist experiment visualization artifacts")
              
    #log the dataframe with the study summary  
    study.trials_dataframe().describe().to_html(experiment.name + ".html")  
    self.mlflow_client.log_artifact(run_id = parent_run.info.run_id, 
                        local_path = experiment.name + ".html")
          
    #log the best hyperparameters in the parent run
    self.mlflow_client.log_metric(parent_run.info.run_id, "loss", study.best_value)
    for k, v in study.best_params.items():
      self.mlflow_client.log_param(parent_run.info.run_id, k, v)

    print('Best trial: {} value: {} params: {}\n'.format(study.best_trial, study.best_value, study.best_params))    



from pytorch_lightning.callbacks import Callback
class PruneRun(Callback):
  #enable run pruning
  def on_batch_end(self, trainer, pl_module):
      if (pl_module.step % 10 == 0):
          this_run = get_existing_run(run = pl_module.run)
          if this_run and not (this_run.info.status == "RUNNING"):
              print(f"Pruning {this_run.info.run_id}({this_run.info.status}) at step {pl_module.step}")
              trainer.should_stop = True



class CircleExperimentModel(pl.LightningModule):
  @staticmethod
  def build_hidden_layers(num_hidden_neurons, activation):

    #create a list of the linear (feedforward) layers
    linear_layers = [ pt.nn.Linear(num_hidden_neurons[i], num_hidden_neurons[i+1]) for i in range(len(num_hidden_neurons) - 1) ]

    #create a list with the required number of the activation functions
    #based on the number of the hidden layers
    classes = [activation.__class__] * len(num_hidden_neurons)
    
    #create a list of the activation function instances
    activation_instances = list(map(lambda x: x(), classes))

    #zip the linear layers with the activation functions 
    hidden_layer_activation_tuples = list(zip(linear_layers, activation_instances))

    #and return as a flat list
    hidden_layers = [i for sublist in hidden_layer_activation_tuples for i in sublist]

    return hidden_layers

  def __init__(self, client, run):
    super().__init__()
    self.start_ts = time.time()
    self.step = 0

    self.client = client
    self.run = run
    self.hparams = run.data.params

    #create a list of hidden layer neurons, e.g. [2, 4, 8]
    num_hidden_neurons = json.loads(self.hparams['num_hidden_neurons'])

    #define the nn model
    self.model = pt.nn.Sequential(
        pt.nn.Linear(2, num_hidden_neurons[0]),
        pt.nn.ReLU(),
        *self.build_hidden_layers(num_hidden_neurons, pt.nn.ReLU()),
        pt.nn.Linear(num_hidden_neurons[-1], 1)
    )

  def forward(self, X):
    y_est = self.model(X)
    return y_est

  def training_step(self, batch, batch_idx):
    batch_size = int( self.run.data.params['batch_size'] ) if 'batch_size' in self.run.data.params else 128
    
    batch = batch.reshape(batch_size, 3)
    y, X = batch[:, 0], batch[:, [1, 2]]
    y_est = self.forward(X).squeeze(dim = 1)#.reshape( (y.shape[0], 1) )
    loss = mse_loss( y_est, y )

    self.client.log_metric(self.run.info.run_id, "loss", loss.item(), step=self.step)
    self.client.log_metric(self.run.info.run_id, "errors", (y_est.round() - y).to(pt.int).sum().abs().item(), step=self.step)
    self.client.log_metric(self.run.info.run_id, "step", self.step, step=self.step)    
    print(f"{self.run.info.run_id}: loss={loss.item()}, step={self.step}")

    self.step += 1
    return loss

  def configure_optimizers(self):
    hparams = self.hparams
    if hparams['optimizer'] == 'Adam':
      lr = float(hparams['lr'])
      optimizer = pt.optim.Adam(self.parameters(), lr = lr)
    else:
      lr = float(hparams['lr'])
      optimizer = pt.optim.SGD(self.parameters() ,lr = lr)

    return optimizer

class CircleExperimentService(BaseOptunaService):
  def hparams(self):
    trial = self._trial

    #define hyperparameters
    return {
      "seed": trial.suggest_int('seed', 0, np.iinfo(np.int32).max - 1),
        
      "optimizer": trial.suggest_categorical('optimizer', ['Adam']),
        
      "lr": trial.suggest_loguniform('lr', 0.061, 0.071),
        
      "num_hidden_neurons": [trial.suggest_categorical(f"num_hidden_layer_{layer}_neurons", [7, 11, 13, 23]) \
                                for layer in range(trial.suggest_categorical('num_layers', [3]))],
      
      "batch_size": trial.suggest_categorical('batch_size', [128]),
        
      "max_batches": trial.suggest_int('max_batches', 150, 250, log = True)
    }

  def on_experiment_end(self, experiment, parent_run):
    study = self._study
    try:
      for key, fig in {
        "plot_param_importances": optuna.visualization.plot_param_importances(study),
        "plot_parallel_coordinate_all": optuna.visualization.plot_parallel_coordinate(study, params=["max_batches", "lr", "num_hidden_layer_0_neurons", "num_hidden_layer_1_neurons", "num_hidden_layer_2_neurons"]),
        "plot_parallel_coordinate_l0_l1_l2": optuna.visualization.plot_parallel_coordinate(study, params=["num_hidden_layer_0_neurons", "num_hidden_layer_1_neurons", "num_hidden_layer_2_neurons"]),
        "plot_contour_max_batches_lr": optuna.visualization.plot_contour(study, params=["max_batches", "lr"]),
      }.items():
        fig.write_image(key + ".png")
        self.mlflow_client.log_artifact(run_id = parent_run.info.run_id, 
                            local_path = key + ".png")
        
    except:
      print(f"Failed to correctly persist experiment visualization artifacts")
      import traceback
      traceback.print_exc()
              
    #log the dataframe with the study summary  
    study.trials_dataframe().describe().to_html(experiment.name + ".html")  
    self.mlflow_client.log_artifact(run_id = parent_run.info.run_id, 
                        local_path = experiment.name + ".html")
          
    #log the best hyperparameters in the parent run
    self.mlflow_client.log_metric(parent_run.info.run_id, "loss", study.best_value)
    for k, v in study.best_params.items():
      self.mlflow_client.log_param(parent_run.info.run_id, k, v)

    print('Best trial: {} value: {} params: {}\n'.format(study.best_trial, study.best_value, study.best_params))    

class CircleExperiment(BaseMLFlowClient, Callback):
    def on_run_start(self, run_idx, run):
        seed = int(run.data.params['seed']) if 'seed' in run.data.params else int( datetime.now().microsecond )
        print(f"{run}({run.info.status}) using seed={seed}")
        np.random.seed(seed)
        pt.manual_seed(seed)

        batch_size = int( run.data.params['batch_size'] )
        train_dl = DataLoader(ObjectStorageDataset(f"s3://dc-taxi-74765ef3897d44692b78bb1532b85354-us-east-2/circle.csv", 
                                        storage_options = {'anon' : False },
                                        eager_load_batches = True,
                                        dtype='float32',
                                        batch_size = batch_size,))

        model = CircleExperimentModel(self.mlflow_client, run)
        max_batches = int( run.data.params['max_batches'] ) if 'max_batches' in run.data.params else 1
        trainer = pl.Trainer(limit_train_batches = max_batches,
                            max_epochs = 1, #set to 1 since the training duration is controlled by max_batches
                            callbacks = [self],
                            progress_bar_refresh_rate = 0, 
                            weights_summary = None,)

        #train the model
        trainer.fit(model, train_dl)

    #enable run pruning
    def on_batch_end(self, trainer, pl_module):
        if (pl_module.step % 10 == 0):
            this_run = get_existing_run(run = pl_module.run)
            if this_run and not (this_run.info.status == "RUNNING"):
                print(f"Pruning {this_run.info.run_id}({this_run.info.status}) at step {pl_module.step}")
                trainer.should_stop = True


from setuptools import setup, find_packages

setup(
	name='laai',
	version='0.0.4.2',
	py_modules=['laai.main'],
	install_requires=['click', 'docker', 'mlflow'],
	extras_require={
        "optuna":  ["optuna", "sklearn", "plotly", "kaleido"],
    },
    include_package_data=True,
	packages = find_packages(),
	entry_points='''
		[console_scripts]
		laai=laai.main:cli
	'''
)




import os
import time
import argparse
import random
from datetime import datetime

from laai.hpo.mlflow_utils import get_client, get_parent_run, get_active_child_run, assert_mlflow_tracking_uri, assert_mlflow_experiment_name

class BaseMLFlowClient():
    def on_run_start(self, run_idx, run):
        self.step = 0
        for i in range(10):
            loss = random.random()
            self.mlflow_client.log_metric(run.info.run_id, "loss", loss, step=i)
            self.mlflow_client.log_metric(run.info.run_id, "step", i, step=i)
            print(f"{run.info.run_id}: loss={loss}, step={i}")
    
    def on_run_terminated(self, run_idx, run):
        pass

    def run(self, num_runs = 1, mlflow_tracking_uri = None, mlflow_experiment_name = None):
        assert num_runs and isinstance(num_runs, int) and num_runs > 0, "The value of num_runs must be a positive integer."
        self.num_runs = num_runs

        #ensure that the mlflow tracking uri is set correctly
        if not mlflow_tracking_uri:
            assert 'MLFLOW_TRACKING_URI' in os.environ, "Specify mlflow_tracking_uri or the MLFLOW_TRACKING_URI environment variable."
            mlflow_tracking_uri = os.environ['MLFLOW_TRACKING_URI']
        self.mlflow_tracking_uri = mlflow_tracking_uri

        #create an mlflow client
        self.mlflow_client = get_client(self.mlflow_tracking_uri)

        #ensure that the mlflow_experiment_name is set correctly
        if not mlflow_experiment_name:
            assert 'MLFLOW_EXPERIMENT_NAME' in os.environ, "Specify mlflow_experiment_name or the MLFLOW_EXPERIMENT_NAME environment variable."  
            mlflow_experiment_name = os.environ['MLFLOW_EXPERIMENT_NAME']
        self.mlflow_experiment_name = mlflow_experiment_name

        #retrieve the experiement
        self.experiment = self.mlflow_client.get_experiment_by_name(mlflow_experiment_name)
        assert self.experiment, f"Unable to retrieve experiment named {mlflow_experiment_name} from {self.mlflow_tracking_uri}."

        parent_run = get_parent_run(self.mlflow_client, 
                                    self.experiment.name, 
                                    self.experiment.experiment_id)

        while num_runs \
                and parent_run \
                and 'RUNNING' == parent_run.info.status:
                    
            active_run = get_active_child_run(client = self.mlflow_client, 
                                                parent_run = parent_run)        
            if active_run:
                print(f"Starting {active_run}({active_run.info.status})")
                try:
                    seed = int(active_run.data.params['seed']) if 'seed' in active_run.data.params else int( datetime.now().microsecond )
                    print(f"{active_run}({active_run.info.status}) using seed={seed}")
                    random.seed(seed)

                    self.on_run_start(self.num_runs - num_runs, active_run)

                finally:
                    self.mlflow_client.set_terminated(active_run.info.run_id)
                    self.on_run_terminated(self.num_runs - num_runs, active_run)
                    print(f"Terminated {active_run.info.run_id}({active_run.info.status})")        

                    num_runs -= 1
                
            parent_run = get_parent_run(self.mlflow_client, self.experiment.name, self.experiment.experiment_id)

        if parent_run:
            print(f"Parent run status: {parent_run.info.status}, terminating...")
        else:
            print(f"Parent run unavailable, terminating...")    

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Start a client for an MLFlow based hyperparameter optimization service.')
    parser.add_argument('--num_runs', default = 1, type = int,
                        help='OPTIONAL: The number of the hyperparameter runs to execute as part of the experiment. Defaults to 1.')
                        
    args = parser.parse_args()

    client = BaseMLFlowClient()
    client.run( num_runs = int(args.num_runs) if isinstance(args.num_runs, int) else 1 )